"""
#=========================================================================
# OpenLoopCLPass.py
#=========================================================================
# Generate a simple schedule (no Mamba techniques here) based on the
# DAG generated by some previous pass.
#
# Author : Shunning Jiang
# Date   : Apr 20, 2019
"""
from __future__ import absolute_import, division, print_function

from graphviz import Digraph

from pymtl3.dsl import CalleePort, NonBlockingCalleeIfc
from pymtl3.dsl.errors import UpblkCyclicError

from .BasePass import BasePass, PassMetadata
from .CLLineTracePass import CLLineTracePass
from .errors import PassOrderError


class OpenLoopCLPass( BasePass ):
  def __call__( self, top ):
    if not hasattr( top._dag, "all_constraints" ):
      raise PassOrderError( "all_constraints" )

    top._sched = PassMetadata()

    self.schedule_with_top_level_callee( top )

  def schedule_with_top_level_callee( self, top ):

    # Construct the graph with top level callee port

    V = top.get_all_update_blocks() | top._dag.genblks

    # We collect all top level callee ports/nonblocking callee interfaces

    top_level_callee_ports = top.get_all_object_filter(
      lambda x: isinstance(x, CalleePort) and x.get_host_component() is top )

    top_level_nb_ifcs = top.get_all_object_filter(
      lambda x: isinstance(x, NonBlockingCalleeIfc) and x.get_host_component() is top )

    # We still tell the top level
    method_callee_mapping = {}
    method_guard_mapping  = {}

    # First deal with normal calleeports. We map the actual method to the
    # callee port, and add the port to the vertex set
    for x in top_level_callee_ports:
      if not x.in_non_blocking_interface(): # Normal callee port
        V.add(x)
        assert x.method not in method_callee_mapping
        method_callee_mapping[x.method] = x

    # Then deal with non-blocking callee interfaces. Map the method of the
    # interface to the actual method and set up method-rdy mapping

    for x in top_level_nb_ifcs:
      V.add( x.method )
      method_guard_mapping [x.method] = x.rdy
      assert x.method.method not in method_callee_mapping
      method_callee_mapping[x.method.method] = x.method

    E   = top._dag.all_constraints
    Es  = { v: [] for v in V }
    InD = { v: 0  for v in V }

    for (u, v) in E: # u -> v
      InD[v] += 1
      Es [u].append( v )

    # In addition to existing constraints, we process the constraints that
    # involve top level callee ports. NOTE THAT we assume the user never
    # set the constraint on the actual method object inside the CalleePort
    # In GenDAGPass we already collect those constraints between update
    # blocks and ACTUAL METHODs. We use the ACTUAL METHOD to callee
    # mapping we set up above to avoid missing constraints.

    for (xx, yy) in top._dag.top_level_callee_constraints:

      if xx in method_callee_mapping:
        xx = method_callee_mapping[ xx ]

      if yy in method_callee_mapping:
        yy = method_callee_mapping[ yy ]

      InD[yy] += 1
      Es [xx].append( yy )
      E.add( (xx, yy) )

    # Perform topological sort for a serial schedule.

    schedule = []

    Q = list( [ v for v in V if not InD[v] ] )

    while Q:
      import random
      random.shuffle(Q)

      # Prioritize update blocks instead of method
      # TODO make it O(logn) by balanced BST if needed ...

      u = None
      for i in range(len(Q)):
        if Q[i] not in method_guard_mapping:
          u = Q.pop(i)
          break

      if u is None:
        u = Q.pop()

      if u in method_guard_mapping:
        schedule.append( method_guard_mapping[ u ] )
      schedule.append( u )
      for v in Es[u]:
        InD[v] -= 1
        if not InD[v]:
          Q.append( v )

    # Shunning: we call CL line trace pass here.
    cl_trace = CLLineTracePass()
    schedule.insert( 0, cl_trace.process_component( top ) )

    top._sched.new_schedule_index  = 0
    top._sched.orig_schedule_index = 0

    # Here we are trying to avoid scanning the original schedule that
    # contains methods because we will need isinstance in that case.
    # As a result we created a preprocessed list for execution and use
    # the dictionary to look up the new index of functions.

    # The last element is always line trace
    def print_line_trace():
      print(top.line_trace())

    schedule.append( print_line_trace )

    schedule_no_method = [ x for x in schedule if not isinstance(x, CalleePort) ]
    mapping = { x : i for i, x in enumerate( schedule_no_method ) }

    def wrap_method( top, method,
                     my_idx_new, next_idx_new, schedule_no_method,
                     my_idx_orig, next_idx_orig ):

      def actual_method( *args, **kwargs ):
        i = top._sched.new_schedule_index
        j = top._sched.orig_schedule_index

        if j > my_idx_orig:
          # This means we need to advance the current cycle to the end
          # and then normally execute until we get to the same point.
          # We use original schedule index to handle the case where
          # there are two consecutive methods.
          while i < len(schedule_no_method):
            schedule_no_method[i]()
            i += 1
          i = j = 0
          top.num_cycles_executed += 1

        # We advance from the current point i to the method's position in
        # the schedule without method just to execute those blocks
        while i < my_idx_new:
          schedule_no_method[i]()
          i += 1

        # Execute the method
        ret = method( *args, **kwargs )

        # Execute all update blocks before the next method. Note that if
        # there are several consecutive methods, my_idx_new is equal to next_idx_new
        while i < next_idx_new:
          schedule_no_method[i]()
          i += 1
        j = next_idx_orig

        if i == len(schedule_no_method):
          i = j = 0
          top.num_cycles_executed += 1

        top._sched.new_schedule_index = i
        top._sched.orig_schedule_index = j
        return ret

      return actual_method

    for i, x in enumerate( schedule ):
      if isinstance( x, CalleePort ):
        x.original_method = x.method

        # This is to find the next non-method block's position in the
        # original schedule
        next_func   = i + 1
        while next_func < len(schedule):
          if not isinstance( schedule[next_func], CalleePort ):
            break
          next_func += 1

        # Get the index of the block in the schedule without method
        # This always exists because we append a line trace at the end
        map_next_func = mapping[ schedule[next_func] ]

        # Get the index of the next method in the schedule without method
        next_method = i + 1
        while next_method < len(schedule):
          if isinstance( schedule[next_method], CalleePort ):
            break
          next_method += 1

        # If there is another method after me, I calculate the range of
        # blocks that I need to call and then stop before the user calls
        # the next method.
        if next_method < len(schedule):
          next_func = next_method
          while next_func < len(schedule):
            if not isinstance( schedule[next_func], CalleePort ):
              break
            next_func += 1
          # Get the index in the compacted schedule
          map_next_func_of_next_method = mapping[ schedule[next_func] ]
        else:
          map_next_func_of_next_method = len(schedule_no_method)

        x.method = wrap_method( top, x.method,
                                map_next_func, map_next_func_of_next_method,
                                schedule_no_method,
                                i, next_method )
    top.num_cycles_executed = 0

    return schedule



"""
#=========================================================================
# LineTraceParamPass_test.py
#=========================================================================
# Test for multi-level line trace pass.
#
# Author : Yanghui Ou
#   Date : 29 May, 2019
"""
from __future__ import absolute_import, division, print_function

from pymtl3.dsl import *
from pymtl3.passes.CLLineTracePass import CLLineTracePass
from pymtl3.passes.GenDAGPass import GenDAGPass
from pymtl3.passes.LineTraceParamPass import LineTraceParamPass
from pymtl3.passes.SimpleSchedulePass import SimpleSchedulePass
from pymtl3.passes.SimpleTickPass import SimpleTickPass
from pymtl3.passes.WrapGreenletPass import WrapGreenletPass


def test_multi_level_trace():

  class Top( Component ):
    def construct( s ):
      s.simple_trace = 'simple'
      s.verbose_trace = 'verbose' 

    def line_trace( s, level='simple' ):
      if level == 'simple':
        return "{}".format( s.simple_trace )
      elif level == 'verbose':
        return "{}".format( s.verbose_trace )
      else:
        return "default"

  A = Top()
  A.set_param( 'top.line_trace', level='verbose' )
  A.elaborate()
  A.apply( GenDAGPass()              )
  A.apply( WrapGreenletPass()        )
  A.apply( SimpleSchedulePass()      )
  A.apply( CLLineTracePass()         )
  A.apply( SimpleTickPass()          )
  A.apply( LineTraceParamPass()      )
  A.lock_in_simulation()
  A.tick()
  print( A.line_trace() )
  assert A.line_trace() == "verbose"



#=========================================================================
# YosysStructuralTranslatorL2_test.py
#=========================================================================
# Author : Peitian Pan
# Date   : June 13, 2019
"""Test the level 2 yosys-SystemVerilog structural translator."""

from __future__ import absolute_import, division, print_function

from pymtl3.passes.rtlir.util.test_utility import do_test
from pymtl3.passes.sverilog.translation.structural.test.SVStructuralTranslatorL1_test import (
    check_eq,
    is_sverilog_reserved,
)
from pymtl3.passes.sverilog.translation.structural.test.SVStructuralTranslatorL2_test import (
    test_nested_struct_port,
    test_packed_array,
    test_struct_const_structural,
    test_struct_packed_array,
    test_struct_port,
)
from pymtl3.passes.yosys.translation.structural.YosysStructuralTranslatorL2 import (
    YosysStructuralTranslatorL2,
)


def local_do_test( m ):
  m.elaborate()
  YosysStructuralTranslatorL2.is_sverilog_reserved = is_sverilog_reserved
  tr = YosysStructuralTranslatorL2( m )
  tr.clear( m )
  tr.translate_structural( m )

  ports = tr.structural.decl_ports[m]
  wires = tr.structural.decl_wires[m]
  conns = tr.structural.connections[m]

  check_eq( ports["port_decls"], m._ref_ports_port_yosys[m] )
  check_eq( ports["wire_decls"], m._ref_ports_wire_yosys[m] )
  check_eq( ports["connections"], m._ref_ports_conn_yosys[m] )
  check_eq( wires, m._ref_wires_yosys[m] )
  check_eq( conns, m._ref_conns_yosys[m] )



#=========================================================================
# utility.py
#=========================================================================
# Author : Peitian Pan
# Date   : Feb 13, 2019
"""Helper methods for RTLIR."""

from __future__ import absolute_import, division, print_function

from functools import reduce


def collect_objs( m, Type ):
  """Return a list of members of `m` that are of type `Type`."""

  def _is_of_type( obj, Type ):
    if isinstance(obj, Type):
      return True
    if isinstance(obj, list):
      return reduce( lambda x, y: x and _is_of_type( y, Type ), obj, True )
    return False

  ret = []
  for name, obj in vars(m).iteritems():
    if isinstance( name, basestring ) and not name.startswith( '_' ):
      if _is_of_type( obj, Type ):
        ret.append( ( name, obj ) )
  return ret



#=========================================================================
# StructuralTranslatorL2_test.py
#=========================================================================
# Author : Peitian Pan
# Date   : May 21, 2019
"""Test the level 2 structural translators."""

from __future__ import absolute_import, division, print_function

import pytest

from pymtl3.datatypes import Bits16, Bits32, BitStruct
from pymtl3.dsl import Component, InPort, OutPort, Wire
from pymtl3.passes.rtlir import RTLIRDataType as rdt
from pymtl3.passes.rtlir.errors import RTLIRConversionError
from pymtl3.passes.rtlir.util.test_utility import do_test, expected_failure
from pymtl3.passes.translator.structural.StructuralTranslatorL2 import (
    StructuralTranslatorL2,
)

from .TestStructuralTranslator import mk_TestStructuralTranslator


def local_do_test( m ):
  m.elaborate()
  tr = mk_TestStructuralTranslator(StructuralTranslatorL2)(m)
  tr.clear( m )
  tr.translate_structural(m)
  try:
    name = tr.structural.component_unique_name[m]
    assert name == m._ref_name
    decl_ports = tr.structural.decl_ports[m]
    assert decl_ports == m._ref_ports
    decl_wires = tr.structural.decl_wires[m]
    assert decl_wires == m._ref_wires
    decl_consts = tr.structural.decl_consts[m]
    assert decl_consts == m._ref_consts
    connections = tr.structural.connections[m]
    assert connections == m._ref_conns
    struct_types = tr.structural.decl_type_struct
    assert sorted(struct_types, key=lambda x: str(x[1])) == m._ref_structs
  except AttributeError:
    pass

def test_struct_port_decl( do_test ):
  class B( BitStruct ):
    def __init__( s, foo=0, bar=42 ):
      s.foo = Bits32(foo)
      s.bar = Bits16(bar)
  class A( Component ):
    def construct( s ):
      s.struct = InPort( B )
  a = A()
  a._ref_name = "A"
  a._ref_ports = \
"""\
port_decls:
  port_decl: struct Port of Struct B
"""
  a._ref_wires = "wire_decls:\n"
  a._ref_consts = "const_decls:\n"
  a._ref_conns = "connections:\n"
  a._ref_structs = [(rdt.Struct('B',
    {'foo':rdt.Vector(32), 'bar':rdt.Vector(16)}, ['bar', 'foo']), 'B')]
  a._ref_src = \
"""\
struct B
component A
(
port_decls:
  port_decl: struct Port of Struct B
interface_decls:
);
const_decls:
freevars:
wire_decls:
component_decls:
tmpvars:
upblk_srcs:
connections:

endcomponent
"""
  do_test( a )

def test_struct_wire_decl( do_test ):
  class B( BitStruct ):
    def __init__( s, foo=0, bar=42 ):
      s.foo = Bits32(foo)
      s.bar = Bits16(bar)
  class A( Component ):
    def construct( s ):
      s.struct = Wire( B )
      @s.update
      def upblk():
        s.struct.foo = 0
        s.struct.bar = Bits16(42)
  a = A()
  a._ref_name = "A"
  a._ref_ports = "port_decls:\n"
  a._ref_wires = \
"""\
wire_decls:
  wire_decl: struct Wire of Struct B
"""
  a._ref_consts = "const_decls:\n"
  a._ref_conns = "connections:\n"
  a._ref_structs = [(rdt.Struct('B',
    {'foo':rdt.Vector(32), 'bar':rdt.Vector(16)}, ['bar', 'foo']), 'B')]
  a._ref_src = \
"""\
struct B
component A
(
port_decls:
interface_decls:
);
const_decls:
freevars:
wire_decls:
  wire_decl: struct Wire of Struct B
component_decls:
tmpvars:
upblk_srcs:
  upblk_src: upblk
connections:

endcomponent
"""
  do_test( a )

@pytest.mark.xfail( reason = "RTLIR not support const struct instance yet" )
def test_struct_const_decl( do_test ):
  class B( BitStruct ):
    def __init__( s, foo=0, bar=42 ):
      s.foo = Bits32(foo)
      s.bar = Bits16(bar)
  class A( Component ):
    def construct( s ):
      s.struct = B()
  a = A()
  a._ref_name = "A"
  a._ref_ports = "port_decls:\n"
  a._ref_wires = "wire_decls:\n"
  a._ref_consts = \
"""\
const_decls:
  const_decl: struct Const of Struct B
"""
  a._ref_conns = "connections:\n"
  a._ref_structs = [(rdt.Struct('B',
    {'foo':rdt.Vector(32), 'bar':rdt.Vector(16)}, ['bar', 'foo']), 'B')]
  a._ref_src = \
"""\
struct B
component A
(
port_decls:
interface_decls:
);
const_decls:
  const_decl: struct Const of Struct B
freevars:
wire_decls:
component_decls:
tmpvars:
upblk_srcs:
connections:

endcomponent
"""
  do_test( a )

def test_struct_port_array( do_test ):
  class B( BitStruct ):
    def __init__( s, foo=0, bar=42 ):
      s.foo = Bits32(foo)
      s.bar = Bits16(bar)
  class A( Component ):
    def construct( s ):
      s.struct = [ InPort( B ) for _ in range(5) ]
  a = A()
  a._ref_name = "A"
  a._ref_ports = \
"""\
port_decls:
  port_decl: struct Array[5] of Port
"""
  a._ref_wires = "wire_decls:\n"
  a._ref_consts = "const_decls:\n"
  a._ref_conns = "connections:\n"
  a._ref_structs = [(rdt.Struct('B',
    {'foo':rdt.Vector(32), 'bar':rdt.Vector(16)}, ['bar', 'foo']), 'B')]
  a._ref_src = \
"""\
struct B
component A
(
port_decls:
  port_decl: struct Array[5] of Port
interface_decls:
);
const_decls:
freevars:
wire_decls:
component_decls:
tmpvars:
upblk_srcs:
connections:

endcomponent
"""
  do_test( a )

def test_nested_struct_port_decl( do_test ):
  class C( BitStruct ):
    def __init__( s, bar=42 ):
      s.bar = Bits16(bar)
  class B( BitStruct ):
    def __init__( s, foo=0 ):
      s.foo = Bits32(foo)
      s.bar = C()
  class A( Component ):
    def construct( s ):
      s.struct = InPort( B )
  a = A()
  a._ref_name = "A"
  a._ref_ports = \
"""\
port_decls:
  port_decl: struct Port of Struct B
"""
  a._ref_wires = "wire_decls:\n"
  a._ref_consts = "const_decls:\n"
  a._ref_conns = "connections:\n"
  a._ref_structs = [(rdt.Struct('B',
    {'foo':rdt.Vector(32),
      'bar':rdt.Struct('C', {'bar':rdt.Vector(16)}, ['bar'])},
    ['bar', 'foo']),
    'B'),
    (rdt.Struct('C', {'bar':rdt.Vector(16)}, ['bar']), 'C')
  ]
  a._ref_src = \
"""\
struct C
struct B
component A
(
port_decls:
  port_decl: struct Port of Struct B
interface_decls:
);
const_decls:
freevars:
wire_decls:
component_decls:
tmpvars:
upblk_srcs:
connections:

endcomponent
"""
  do_test( a )

def test_struct_packed_array_port_decl( do_test ):
  class B( BitStruct ):
    def __init__( s, foo=0, bar=42 ):
      s.foo = [ Bits32(foo) for _ in range( 5 ) ]
      s.bar = Bits16(bar)
  class A( Component ):
    def construct( s ):
      s.struct = InPort( B )
  a = A()
  a._ref_name = "A"
  a._ref_ports = \
"""\
port_decls:
  port_decl: struct Port of Struct B
"""
  a._ref_wires = "wire_decls:\n"
  a._ref_consts = "const_decls:\n"
  a._ref_conns = "connections:\n"
  a._ref_structs = [(rdt.Struct('B',
    {'foo':rdt.PackedArray([5], rdt.Vector(32)),
     'bar':rdt.Vector(16) },
    ['bar', 'foo']),
    'B')
  ]
  a._ref_src = \
"""\
struct B
component A
(
port_decls:
  port_decl: struct Port of Struct B
interface_decls:
);
const_decls:
freevars:
wire_decls:
component_decls:
tmpvars:
upblk_srcs:
connections:

endcomponent
"""
  do_test( a )

def test_nested_struct_packed_array_port_decl( do_test ):
  class C( BitStruct ):
    def __init__( s, bar=42 ):
      s.bar = Bits16(bar)
  class B( BitStruct ):
    def __init__( s, foo=0, bar=42 ):
      s.foo = Bits32(foo)
      s.bar = [ C() for _ in range(5) ]
  class A( Component ):
    def construct( s ):
      s.struct = InPort( B )
  a = A()
  a._ref_name = "A"
  a._ref_ports = \
"""\
port_decls:
  port_decl: struct Port of Struct B
"""
  a._ref_wires = "wire_decls:\n"
  a._ref_consts = "const_decls:\n"
  a._ref_conns = "connections:\n"
  a._ref_structs = [(rdt.Struct('B',
    {'foo':rdt.Vector(32),
     'bar':rdt.PackedArray([5], rdt.Struct('C', {'bar':rdt.Vector(16)}, ['bar']))},
    ['bar', 'foo']),
    'B'),
    (rdt.Struct('C', {'bar':rdt.Vector(16)}, ['bar']), 'C')
  ]
  a._ref_src = \
"""\
struct C
struct B
component A
(
port_decls:
  port_decl: struct Port of Struct B
interface_decls:
);
const_decls:
freevars:
wire_decls:
component_decls:
tmpvars:
upblk_srcs:
connections:

endcomponent
"""
  do_test( a )

def test_nested_struct_packed_array_index( do_test ):
  class C( BitStruct ):
    def __init__( s, bar=42 ):
      s.bar = Bits16(bar)
  class B( BitStruct ):
    def __init__( s, foo=0, bar=42 ):
      s.foo = Bits32(foo)
      s.bar = [ C() for _ in range(5) ]
  class A( Component ):
    def construct( s ):
      s.struct = InPort( B )
      s.out = OutPort( C )
      s.connect( s.struct.bar[1], s.out )
  a = A()
  a._ref_name = "A"
  a._ref_ports = \
"""\
port_decls:
  port_decl: out Port of Struct C
  port_decl: struct Port of Struct B
"""
  a._ref_wires = "wire_decls:\n"
  a._ref_consts = "const_decls:\n"
  a._ref_conns = \
"""\
connections:
  connection: PackedIndex StructAttr CurCompAttr struct bar 1 -> CurCompAttr out
"""
  a._ref_structs = [(rdt.Struct('B',
    {'foo':rdt.Vector(32),
     'bar':rdt.PackedArray([5], rdt.Struct('C', {'bar':rdt.Vector(16)}, ['bar']))},
    ['bar', 'foo']),
    'B'),
    (rdt.Struct('C', {'bar':rdt.Vector(16)}, ['bar']), 'C')
  ]
  a._ref_src = \
"""\
struct C
struct B
component A
(
port_decls:
  port_decl: out Port of Struct C
  port_decl: struct Port of Struct B
interface_decls:
);
const_decls:
freevars:
wire_decls:
component_decls:
tmpvars:
upblk_srcs:
connections:
  connection: PackedIndex StructAttr CurCompAttr struct bar 1 -> CurCompAttr out

endcomponent
"""
  do_test( a )

__all__ = list(filter(lambda s: s.startswith('test_'), dir()))



"""
========================================================================
HeuristicTopo.py
========================================================================

Author : Shunning Jiang
Date   : Dec 26, 2018
"""

from __future__ import absolute_import, division, print_function

import ast

from graphviz import Digraph

from pymtl3.passes.BasePass import BasePass, PassMetadata
from pymtl3.passes.errors import PassOrderError
from pymtl3.passes.SimpleSchedulePass import check_schedule

try:
  from Queue import PriorityQueue
except ImportError:
  from queue import PriorityQueue


class CountBranches( ast.NodeVisitor ):

  def enter( self, node ):
    self.num_br = 0
    self.visit( node )
    return self.num_br

  def visit_If( self, node ):
    self.num_br += 1
    self.visit( node.test )

    for stmt in node.body:
      self.visit( stmt )

    for stmt in node.orelse:
      self.visit( stmt )

  def visit_IfExp( self, node ):
    self.num_br += 1
    self.visit( node.test )
    self.visit( node.body )
    self.visit( node.orelse )

  # A single for is estimated to be 10x of a branch
  def visit_For( self, node ):
    self.num_br += 0
    for stmt in node.body:
      self.visit( stmt )

  def visit_While( self, node ):
    self.num_br += 0
    for stmt in node.body:
      self.visit( stmt )

class HeuristicTopoPass( BasePass ):
  def __call__( self, top ):
    if not hasattr( top._dag, "all_constraints" ):
      raise PassOrderError( "all_constraints" )

    top._sched = PassMetadata()

    top._sched.schedule = self.schedule( top )

  def schedule( self, top ):

    # Construct the graph

    V   = top._dag.final_upblks
    E   = top._dag.all_constraints
    Es  = { v: [] for v in V }
    InD = { v: 0  for v in V }

    for (u, v) in E: # u -> v
      InD[v] += 1
      Es [u].append( v )

    # Extract branchiness

    # Initialize all generated net block to 0 branchiness
    branchiness = { x: 0 for x in net_blks }

    visitor = CountBranches()
    for blk in normal_blks:
      hostobj = top.get_update_block_host_component( blk )
      branchiness[ blk ] = visitor.enter( hostobj.get_update_block_ast( blk ) )

    # Perform topological sort for a serial schedule.
    # Note that here we use a priority queue to get the blocks with small
    # branchiness as early as possible

    schedule = []

    Q = PriorityQueue(0)
    for v in V:
      if not InD[v]:
        Q.put( (branchiness[ v ], v) )

    while not Q.empty():
      br, u = Q.get()
      schedule.append( u )
      for v in Es[u]:
        InD[v] -= 1
        if not InD[v]:
          Q.put( (branchiness[ v ], v) )

    check_schedule( top, schedule, V, E, InD )

    return schedule



#=========================================================================
# BehavioralRTLIRTypeCheckL5Pass.py
#=========================================================================
# Author : Peitian Pan
# Date   : March 30, 2019
"""Provide L5 behavioral RTLIR type check pass."""
from __future__ import absolute_import, division, print_function

from collections import OrderedDict

from pymtl3.passes.BasePass import BasePass, PassMetadata
from pymtl3.passes.rtlir.errors import PyMTLTypeError
from pymtl3.passes.rtlir.rtype import RTLIRType as rt

from .BehavioralRTLIRTypeCheckL4Pass import BehavioralRTLIRTypeCheckVisitorL4


class BehavioralRTLIRTypeCheckL5Pass( BasePass ):
  def __call__( s, m ):
    """Perform type checking on all RTLIR in rtlir_upblks."""
    if not hasattr( m, '_pass_behavioral_rtlir_type_check' ):
      m._pass_behavioral_rtlir_type_check = PassMetadata()
    m._pass_behavioral_rtlir_type_check.rtlir_freevars = OrderedDict()
    m._pass_behavioral_rtlir_type_check.rtlir_tmpvars = OrderedDict()
    m._pass_behavioral_rtlir_type_check.rtlir_accessed = set()

    visitor = BehavioralRTLIRTypeCheckVisitorL5( m,
      m._pass_behavioral_rtlir_type_check.rtlir_freevars,
      m._pass_behavioral_rtlir_type_check.rtlir_accessed,
      m._pass_behavioral_rtlir_type_check.rtlir_tmpvars
    )

    for blk in m.get_update_blocks():
      visitor.enter( blk, m._pass_behavioral_rtlir_gen.rtlir_upblks[ blk ] )

class BehavioralRTLIRTypeCheckVisitorL5( BehavioralRTLIRTypeCheckVisitorL4 ):
  def __init__( s, component, freevars, accessed, tmpvars ):
    super( BehavioralRTLIRTypeCheckVisitorL5, s ). \
        __init__( component, freevars, accessed, tmpvars )

  def visit_Index( s, node ):
    """Type check the index node.

    Only static constant expressions can be the index of component arrays
    """
    if isinstance( node.value.Type, rt.Array ) and \
       isinstance( node.value.Type.get_sub_type(), rt.Component ):
      node.Type = node.value.Type.get_sub_type()

    else:
      super( BehavioralRTLIRTypeCheckVisitorL5, s ).visit_Index( node )

  def visit_Attribute( s, node ):
    """Type check an attribute.

    Since only the ports of a subcomponent can be accessed, no explicit
    cross-hierarchy access detection is needed.
    """
    # Attributes of subcomponent can only access ports
    if isinstance( node.value.Type, rt.Component ) and \
       node.value.Type.get_name() != s.component.__class__.__name__:
      if not node.value.Type.has_property( node.attr ):
        raise PyMTLTypeError( s.blk, node.ast,
          'rt.Component {} does not have attribute {}!'.format(
            node.value.Type.get_name(), node.attr ) )
      prop = node.value.Type.get_property( node.attr )
      if not rt._is_of_type( prop, ( rt.Port, rt.InterfaceView ) ):
        raise PyMTLTypeError( s.blk, node.ast,
          '{} is not a port of {} subcomponent!'.format(
            node.attr, node.value.Type.get_name() ) )
      node.Type = prop
    else:
      super( BehavioralRTLIRTypeCheckVisitorL5, s ).visit_Attribute( node )



from __future__ import absolute_import, division, print_function

from .bits_import import *
from .bits_import import _bitwidths
from .BitStruct import BitStruct, mk_bit_struct
from .helpers import clog2, concat, reduce_and, reduce_or, reduce_xor, sext, zext



"""
========================================================================
errors.py
========================================================================

Author : Shunning Jiang
Date   : Jul 4, 2017
"""
from __future__ import absolute_import, division, print_function


class PassOrderError( Exception ):
  """ Raise when applying a pass to a component and some required variable
      generated by other passes is missing """
  def __init__( self, var ):
    return super( PassOrderError, self ).__init__( \
    "Please first apply other passes to generate model.{}".format( var ) )

class ModelTypeError( Exception ):
  """ Raise when a pass cannot be applied to some component type """
  def __init__( self, typename ):
    return super( ModelTypeError, self ).__init__( \
    "This pass can only be applied to {}".format( typename ) )

class TranslationError( Exception ):
  """ Raise translation goes wrong """
  def __init__( self, blk, x ):
    return super( TranslationError, self ).__init__( \
    "{} {}".format( blk.__name__, x ) )



#=========================================================================
# errors.py
#=========================================================================
# Author : Peitian Pan
# Date   : May 27, 2019
"""Exception classes for SystemVerilog backend."""

from __future__ import absolute_import, division, print_function

import inspect
import os
import sys
import traceback


class SVerilogImportError( Exception ):
  """Error while performing import."""
  def __init__( self, obj, msg ):
    obj = str(obj)
    _, _, tb = sys.exc_info()
    traceback.print_tb(tb)
    tb_info = traceback.extract_tb(tb)
    fname, line, func, text = tb_info[-1]
    return super( SVerilogImportError, self ).__init__(
      "\nIn file {fname}, Line {line}, Method {func}:"
      "\nError trying to perform import on {obj}:\n- {msg}"
      "\n  {text}".format( **locals() ) )

class SVerilogTranslationError( Exception ):
  """SystemVerilog translation error."""
  def __init__( self, blk, _ast, msg ):
    ast = _ast.ast
    fname = os.path.abspath( inspect.getsourcefile( blk ) )
    line = inspect.getsourcelines( blk )[1]
    col = 0
    code = ""
    try:
      line += ast.lineno - 1
      col = ast.col_offset
      code_line = inspect.getsourcelines( blk )[0][ ast.lineno-1 ]
      code = '\n  ' + code_line.strip() + \
        '\n  '+ ' ' * (col-len(code_line)+len(code_line.lstrip())) + '^'
    except AttributeError:
      # The given AST node is neither expr nor stmt
      pass
    return super( SVerilogTranslationError, self ).__init__(
      "\nIn file {fname}, Line {line}, Col {col}:{code}\n- {msg}". \
      format( **locals() ) )

class SVerilogReservedKeywordError( Exception ):
  """SystemVerilog reserved keyword error."""
  def __init__( self, name, msg ):
    return super( SVerilogReservedKeywordError, self ).__init__(
      "- {name} is a SystemVerilog reserved keyword!\n- {msg}". \
          format( **locals() ) )

class SVerilogCloseLoopPurePythonSimError( Exception ):
  """SystemVerilog closed loop test error during pure python simulation."""
  def __init__( self, obj, msg ):
    return super( SVerilogCloseLoopPurePythonSimError, self ).__init__(
      "\nCalling closed-loop test method with component {obj}:\n- {msg}". \
          format( **locals() ) )



#=========================================================================
# YosysBehavioralTranslatorL2_test.py
#=========================================================================
# Author : Peitian Pan
# Date   : June 9, 2019
"""Test the SystemVerilog translator implementation."""

from __future__ import absolute_import, division, print_function

from pymtl3.passes.rtlir import BehavioralRTLIRGenPass, BehavioralRTLIRTypeCheckPass
from pymtl3.passes.rtlir.util.test_utility import do_test
from pymtl3.passes.sverilog.translation.behavioral.test.SVBehavioralTranslatorL1_test import (
    is_sverilog_reserved,
)
from pymtl3.passes.sverilog.translation.behavioral.test.SVBehavioralTranslatorL2_test import (
    test_for_range_lower_upper,
    test_for_range_lower_upper_step,
    test_for_range_upper,
    test_if,
    test_if_bool_op,
    test_if_branches,
    test_if_dangling_else_inner,
    test_if_dangling_else_outter,
    test_if_exp_for,
    test_if_exp_unary_op,
    test_nested_if,
    test_reduce,
    test_tmpvar,
)

from ..YosysBehavioralTranslatorL2 import YosysBehavioralRTLIRToSVVisitorL2


def local_do_test( m ):
  m.elaborate()
  m.apply( BehavioralRTLIRGenPass() )
  m.apply( BehavioralRTLIRTypeCheckPass() )

  visitor = YosysBehavioralRTLIRToSVVisitorL2(is_sverilog_reserved)
  upblks = m._pass_behavioral_rtlir_gen.rtlir_upblks
  m_all_upblks = m.get_update_blocks()
  for blk in m_all_upblks:
    upblk_src = visitor.enter( blk, upblks[blk] )
    upblk_src = "\n".join( upblk_src )
    assert upblk_src == m._ref_upblk_srcs_yosys[blk.__name__]
